/*
 * (c) 1999, David Leonard. <david.leonard@csee.uq.edu.au>
 * $Id: gw.c,v 1.1.1.1 1999/10/16 07:19:54 leonard Exp $
 */
/*
 * Top-level pty-telnet bridging layer.
 */
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include "layer.h"
#include "gw.h"
#include "pty.h"
#include "tn.h"

#define structoffield(p,type,memb) (type *)((caddr_t)(p) - offsetof(type,memb))

struct gwlayer;

static void gw_reset(struct gwlayer *);
static void gw_msg(struct gwlayer *, char *);
static void gw_prompt(struct gwlayer *);
static void gw_edit(struct gwlayer *, char);
static void gw_command(struct gwlayer *, char *);
static void gw_putc(struct gwlayer *, char);
static void gw_backspace(struct gwlayer *);

static void gw_tn_recv(struct layer *, void *, size_t);
static void gw_pty_recv(struct layer *, void *, size_t);

static void gw_delete(struct layer *);
static void gw_selset(struct layer *, int *, fd_set *);
static void gw_selrdy(struct layer *, fd_set *);

static struct layerclass gw_class = {
	"at",
	no_recv,
	no_send,
	gw_delete,
	gw_selset,
	gw_selrdy
};

static struct layerclass gw_pty_class = {
	"gw_pty",
	gw_pty_recv,
	below_send,
	below_delete,
	below_selset,
	below_selrdy
};

static struct layerclass gw_tn_class = {
	"gw_tn",
	gw_tn_recv,
	below_send,
	below_delete,
	below_selset,
	below_selrdy
};

struct gwlayer {
	struct layer l;
	struct layer pty;
	struct layer tn;
	char cmdbuf[1024];
	int cmdlen;
	int mode;
#define MODE_EDIT	0
#define MODE_CONNECTED	1
	int echo;
	char *prompt;
};

static void
gw_command(g, cmd)
	struct gwlayer *g;
	char *cmd;
{
	struct in_addr addr;
	in_port_t port;
	struct layer *tn;
	char buf[] = "255.255.255.255:32768";
	char *hp;

#ifdef DEBUG
	fprintf(stderr, "got command: %s\n", cmd);
#endif

	/* skip leading whitespace */
	while (*cmd && isspace(*cmd))
		cmd++;

	if (*cmd == '\0') {
		return;
	} 
	if (strcmp(cmd, "quit") == 0) {
		gw_msg(g, "bye\r\n");
		gw_reset(g);
		return;
	} 

	if (strncmp(cmd, "telnet ", strlen("telnet ")) == 0) {
		hp = cmd + strlen("telnet ");
		while (*hp && isspace(*hp))
			hp++;
		if (!*hp) {
			gw_msg(g, "bad hostname\r\n");
			return;
		}
		    
		if (parse_hostport(hp, &addr, &port))  {
			gw_msg(g, "bad hostname or port\r\n");
			return;
		}

		gw_msg(g, "connecting to ");
		snprintf(buf, sizeof buf, "%s:%u",
			inet_ntoa(addr), port);
		gw_msg(g, buf);
		gw_msg(g, "... ");
		tn = tn_newlayer(&addr, port);
		if (!tn) {
			gw_msg(g, "failed: ");
			gw_msg(g, strerror(errno));
			gw_msg(g, "\r\n");
			return;
		}
		new_below(&g->tn, tn);
		g->mode = MODE_CONNECTED;
		gw_msg(g, "connected\r\n");
		return;
	} 
	gw_msg(g, "bad command\r\n");
}

static void
gw_msg(g, m)
	struct gwlayer *g;
	char *m;
{
	layer_send(&g->pty, m, strlen(m));
}

static void
gw_prompt(g)
	struct gwlayer *g;
{
	gw_msg(g, g->prompt);
}

static void
gw_putc(g, c)
	struct gwlayer *g;
	char c;
{
	layer_send(&g->pty, &c, sizeof c);
}

static void
gw_backspace(g)
	struct gwlayer *g;
{
	g->cmdlen--;
	if (g->echo) {
		if (g->cmdbuf[g->cmdlen] < ' ')
			gw_msg(g, "\b\b  \b\b");
		else
			gw_msg(g, "\b \b");
	}
}

static void
gw_edit(g, c)
	struct gwlayer *g;
	char c;
{
	c &= ~0x80;
	switch (c) {
	case '\n':
	case '\r':
		gw_putc(g, '\r');
		gw_putc(g, '\n');
		g->cmdbuf[g->cmdlen] = '\0';
		gw_command(g, g->cmdbuf);
		g->cmdlen = 0;
		if (g->mode == MODE_EDIT)
			gw_prompt(g);
		break;
	case 0x7f:
	case '\b':
		if (g->cmdlen)
			gw_backspace(g);
		break;
	case 'U' & 0x3f:
		while (g->cmdlen)
			gw_backspace(g);
		break;
	case 'C' & 0x3f:
		gw_msg(g, "[intr]\r\n");
		gw_prompt(g);
		g->cmdlen = 0;
		break;
	default:
		if (g->cmdlen < sizeof g->cmdbuf - 1) {
			g->cmdbuf[g->cmdlen++] = c;
			if (g->echo) {
				if (c < ' ') {
					gw_putc(g, '^');
					gw_putc(g, '@' | c);
				} else 
					gw_putc(g, c);
			}
		}
	}
}

static void
gw_reset(g)
	struct gwlayer *g;
{
	if (g->tn.below)
		layer_delete(g->tn.below);
	g->tn.below = NULL;
	if (g->pty.below)
		pty_reopen(g->pty.below);
	g->cmdlen = 0;
	g->mode = MODE_EDIT;
	gw_prompt(g);
}

struct layer *
gw_newlayer(ptypath, prompt)
	const char *ptypath;
	char *prompt;
{
	struct gwlayer *g;
	struct layer *p;

	p = pty_newlayer(ptypath);
	if (p == NULL)
		return NULL;

	g = (struct gwlayer *)malloc(sizeof *g);
	g->l.class = &gw_class;

	g->pty.class = &gw_pty_class;
	g->pty.below = NULL;

	g->tn.class = &gw_tn_class;
	g->tn.below = NULL;

	g->prompt = prompt;
	g->echo = 1;

	gw_reset(g);

	new_below(&g->pty, p);

	return (struct layer *)g;
}

static void
gw_pty_recv(l, buf, size)
	struct layer *l;
	void *buf;
	size_t size;
{
	struct gwlayer *g = structoffield(l, struct gwlayer, pty);
	unsigned char *c = (unsigned char *)buf;

        if (size == 0) 
		gw_reset(g);
	else
		switch (g->mode) {
		case MODE_EDIT:
			while (g->mode == MODE_EDIT && size > 0) {
			        gw_edit(g, *c++);
			        size--;
			}
			if (size == 0 || g->mode != MODE_CONNECTED)
				break;
		case MODE_CONNECTED:
			if (g->tn.below)
				layer_send(g->tn.below, c, size);
			break;
		}
}

static void
gw_tn_recv(l, buf, size)
	struct layer *l;
	void *buf;
	size_t size;
{
	struct gwlayer *g = structoffield(l, struct gwlayer, tn);

	if (size == 0)  {
		gw_msg(g, "\r\nconnection closed by peer\r\n");
		gw_reset(g);
	} else if (g->pty.below)
		layer_send(g->pty.below, buf, size);
}

static void
gw_selset(l, n, r)
	struct layer *l;
	int *n;
	fd_set *r;
{
	struct gwlayer *g = (struct gwlayer *)l;

	layer_selset(&g->tn, n, r);
	layer_selset(&g->pty, n, r);
}

static void
gw_selrdy(l, r)
	struct layer *l;
	fd_set *r;
{
	struct gwlayer *g = (struct gwlayer *)l;

	layer_selrdy(&g->tn, r);
	layer_selrdy(&g->pty, r);
}

static void
gw_delete(l)
	struct layer *l;
{
	struct gwlayer *g = (struct gwlayer *)l;

	layer_delete(&g->tn);
	layer_delete(&g->pty);
	free(g);
}
