/*
 * (c) 1999, David Leonard. <david.leonard@csee.uq.edu.au>
 * $Id: layer.h,v 1.1.1.1 1999/10/16 07:19:54 leonard Exp $
 */

/*
 * A layer communicates with the layer below by calling its send()
 * method. A layer communicates with the layer above by calling its
 * recv() method. Two other methods (selset, selrdy) are used to
 * set up fd_sets for select() and then to act upon fds that are ready
 * for read. In this way, bottom layers can generate recv()s.
 */

#include <sys/types.h>

struct layer;

struct layerclass {
	const char* name;
	void	(*recv)(struct layer *, void *, size_t); /* from below */
	void	(*send)(struct layer *, void *, size_t); /* from above */
	void	(*delete)(struct layer *);
	void	(*selset)(struct layer *, int *, fd_set *);
	void	(*selrdy)(struct layer *, fd_set *);
};

#ifdef DEBUG
void layer_recv(struct layer *, void *, size_t);
void layer_send(struct layer *, void *, size_t);
void layer_delete(struct layer *);
void layer_selset(struct layer *, int *, fd_set *);
void layer_selrdy(struct layer *, fd_set *);
#else
#define layer_recv(l,b,s)	(*(l)->class->recv)(l,b,s)
#define layer_send(l,b,s)	(*(l)->class->send)(l,b,s)
#define layer_delete(l)		(*(l)->class->delete)(l)
#define layer_selset(l,n,r)	(*(l)->class->selset)(l,n,r)
#define layer_selrdy(l,r)	(*(l)->class->selrdy)(l,r)
#endif

void no_recv(struct layer *, void *, size_t);
void no_send(struct layer *, void *, size_t);
void no_delete(struct layer *);
void no_selset(struct layer *, int *, fd_set *);
void no_selrdy(struct layer *, fd_set *);
void above_recv(struct layer *, void *, size_t);
void below_send(struct layer *, void *, size_t);
void below_selset(struct layer *, int *, fd_set *);
void below_selrdy(struct layer *, fd_set *);
void below_delete(struct layer *);

#define new_below(l,b)	((l)->below = (b))->above = (l)

/*
 * recv(l, buf, size):
 *	- buf may be modified
 *	- size of 0 means that the lower layer is closing
 * send(l, buf, size)
 *	- buf may be modified
 * delete(l)
 *	- l will never be dereferenced again
 * selset(l, n, r)
 *	- r and n should be modified with selset_add()
 * selrdy(l, r)
 *	- r should be checked with FD_ISSET and acted apon as
 *	  it will be ready for read.
 */

#define selset_add(n,r,fd) \
	do { FD_SET(fd, r); if (*(n) < (fd) + 1) { *(n) = (fd) + 1;} } while (0)

struct layer {
	struct layerclass *class;
	struct layer *above;		/* layer to recv upwards */
	struct layer *below;		/* only used by below macros */
};

void layer_select(struct layer *);
