/*
 * (c) 1999, David Leonard. <david.leonard@csee.uq.edu.au>
 * $Id: tn.c,v 1.1.1.1 1999/10/16 07:19:54 leonard Exp $
 */
/*
 * Telnet protocol layer.
 */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include <string.h>
#include <ctype.h>
#include <err.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>

#ifdef DEBUG
#define TELOPTS
#define TELCMDS
#endif
#include <arpa/telnet.h>
#include <arpa/inet.h>

#include "layer.h"
#include "tn.h"
#include "fd.h"

struct tnlayer;

static void	tn_recv(struct layer *, void *, size_t);
static void	tn_send(struct layer *, void *, size_t);
static void	tn_delete(struct layer *);

static void	tn_sendcmd(struct tnlayer *, unsigned char, unsigned char);
static void	tn_recvcmd(struct tnlayer *, unsigned char, unsigned char);

static struct layerclass tn_class = {
	"tn",
	tn_recv,
	tn_send,
	tn_delete,
	below_selset,
	below_selrdy
};

struct tnlayer  {
	struct layer l;
	unsigned char cmd;
	char	dexp[64];	/* DO/DONT expected */
	char	wexp[64];	/* WILL/WONT expected */
};

struct layer *
tn_newlayer(host, port)
	struct in_addr *host;
	in_port_t port;
{
	int fd;
	struct sockaddr_in to, from;
	struct tnlayer *t;
	struct layer *fdl;

	from.sin_len = sizeof from;
	from.sin_family = AF_INET;
	from.sin_port = htons(0);
	from.sin_addr.s_addr = INADDR_ANY;

	to.sin_len = sizeof to;
	to.sin_family = AF_INET;
	to.sin_port = htons(port);
	to.sin_addr = *host;

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		return NULL;
	
	if (bind(fd, (struct sockaddr *)&from, sizeof from) == -1) {
		close(fd);
		return NULL;
	}

	if (connect(fd, (struct sockaddr *)&to, sizeof to) == -1) {
		close(fd);
		return NULL;
	}

#ifdef DEBUG
	fprintf(stderr, "socket connected fd=%d\n", fd);
#endif

	fdl = fd_newlayer(fd);
	if (fdl == NULL) {
		close(fd);
		return NULL;
	}

	t = (struct tnlayer *)malloc(sizeof *t);
	t->l.class = &tn_class;
	t->l.above = NULL;

	t->cmd = 0;
	memset(t->dexp, 0, sizeof t->dexp);
	memset(t->wexp, 0, sizeof t->wexp);

	new_below((struct layer *)t, fdl);

	return (struct layer *)t;
}

/* Close the telnet connection. */
static void
tn_delete(l)
	struct layer *l;
{
	struct tnlayer *t = (struct tnlayer *)l;

	below_delete(l);
	free(t);
}

/* Send a telnet private command. */
static void
tn_sendcmd(t, cmd, op)
	struct tnlayer *t;
	unsigned char cmd;
	unsigned char op;
{
	unsigned char cmdbuf[3] = { IAC, cmd, op };

#ifdef DEBUG
	fprintf(stderr, "tn sendcmd %s %s\n", TELCMD(cmd), TELOPT(op));
#endif
	layer_send(t->l.below, cmdbuf, sizeof cmdbuf);
}

/* Process a received telnet command. */
static void
tn_recvcmd(t, cmd, op)
	struct tnlayer *t;
	unsigned char cmd;
	unsigned char op;
{
	unsigned char ok = 
		(cmd == WILL) ? DO :
		(cmd == WONT) ? DONT :
		(cmd == DO)   ? WILL :
		     /* DONT */ WONT ;

#ifdef DEBUG
	fprintf(stderr, "tn recvcmd %s %s\n", TELCMD(cmd), TELOPT(op));
#endif
	/* Ignore illegal ops */
	if (op > sizeof t->wexp / sizeof t->wexp[0])
		return;

	/* Absorb expected replies */
	if ((cmd == WILL || cmd == WONT) && t->wexp[op]) {
		t->wexp[op] = 0;
		return;
	} else if ((cmd == DO || cmd == DONT) && t->dexp[op]) {
		t->dexp[op] = 0;
		return;
	} else if (cmd == WILL || cmd == WONT)
		switch (op) {
		case TELOPT_BINARY:
		case TELOPT_LINEMODE:
		case TELOPT_ECHO:
			tn_sendcmd(t, ok, op);
			return;
		case TELOPT_SGA:
			tn_sendcmd(t, DO, op);
			return;
		default:
			tn_sendcmd(t, DONT, op);
			return;
		}
	else /* (cmd == DO || cmd == DONT) */
		switch (op) {
		case TELOPT_SGA:
		case TELOPT_BINARY:
			tn_sendcmd(t, ok, op);
			return;
		default:
			tn_sendcmd(t, WONT, op);
			return;
		}
}

/* Receive user data from the fd layer and pass up */
void
tn_recv(l, buf, len)
	struct layer *l;
	void *buf;
	size_t len;
{
	struct tnlayer *t = (struct tnlayer *)l;
	unsigned char *c = (unsigned char *)buf; 
	int in, out;

	if (len == 0) {
		/* close message */
		layer_recv(l->above, buf, len);
		return;
	}

	for (in = out = 0; in < len; in++)
	    switch (t->cmd) {
	    case 0:
		if (c[in] == IAC)
			t->cmd = IAC;
		else
			c[out++] = c[in];
		break;
	    case IAC:
		switch (c[in]) {
		case IAC:
			c[out++] = IAC;
			t->cmd = 0;
			break;
		case WILL:
		case WONT:
		case DO:
		case DONT:
			t->cmd = c[in];
			break;
		}
		break;
	    case WILL:
	    case WONT:
	    case DO:
	    case DONT:
		(void) tn_recvcmd(t, t->cmd, c[in]);
		t->cmd = 0;
		break;
	    }

	/* Only pass up if we have something. */
	if (out)
		layer_recv(l->above, buf, out);
}

void
tn_send(l, buf, len)
	struct layer *l;
	void *buf;
	size_t len;
{
	unsigned char iac[2];
	int start, end;
	unsigned char *c = (unsigned char *)buf;

	for (start = end = 0; end < len; end++) 
		if (c[end] == IAC) {
			if (end > start)
			    layer_send(l->below, c + start, end - start);
			iac[0] = IAC;
			iac[1] = IAC;
			layer_send(l->below, iac, sizeof iac);
			start = end + 1;
		}
	if (end > start)
	    layer_send(l->below, c + start, end - start);
}

int
parse_hostport(hp, a, p)
	char *hp;
	struct in_addr *a;
	in_port_t *p;
{
	char *portstr; 
	char *hostname;
	struct hostent *he;
	struct servent *se;

#ifdef DEBUG
	fprintf(stderr, "parse_hostport(%s)\n", hp);
#endif
	hostname = strtok(hp, " ");
	portstr = strtok(NULL, " ");
	if (!portstr)
		portstr = "telnet";
#ifdef DEBUG
	fprintf(stderr, "hostname=%s portstr=%s\n", hostname, portstr);
#endif
	if (isdigit(*portstr))
		*p = atoi(portstr);
	else {
		se = getservbyname(portstr, "tcp");
		if (!se) {
			warn("getservbyname(%s/tcp)", portstr);
			return -1;
		}
		*p = ntohs(se->s_port);
	}
	he = gethostbyname(hostname);
	if (!he) {
		herror(hostname);
		return -1;
	}
	memcpy(a, (void *)he->h_addr, sizeof *a);
	return 0;
}
